// Copyright (c) 2020 Morel Bérenger
// 
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.
// 
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
// 
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would be
//    appreciated but is not required.
// 2. Altered source versions must be plainly marked as such, and must not be
//    misrepresented as being the original software.
// 3. This notice may not be removed or altered from any source distribution.

// This file defines unique_res and weak_res class utilities.
// goal: easy RAII for any resource, without any assumption
//   about what is an invalid value, and without consuming
//   more data memory than a single, raw pointer.

#ifndef MEMORY_HPP
#define MEMORY_HPP

#include <utility>

// ease the life for memory stuff, just use:
// typedef unique_res<T,nullptr,generic_free<T>> T_owner;
// typedef weak_res<T,nullptr> T_owned;
template <typename T>
void generic_free( T* p )
{
	delete p;
}

// those classes are purely for semantics, and do not
// provide any real guarantee, as anyone is free to call
// `close()`, `free()` or whatever function able to release
// resource in `T`, but at least it should make it easier to
// understand that it is a bad idea.
// Arrays are explicitly not handled, because that's what
// vector is for. If you can't affort the small overhead of
// one pointer per collection, you likely can afford to
// write your own stuff.

template
<
	typename T,
	T INVALID_VAL
>
class weak_res
{
protected:
	typedef weak_res<T,INVALID_VAL> COMPLETE_TYPE;
	T m_data;
public:
	bool valid( void ) const
	{
		return m_data != INVALID_VAL;
	}

	bool operator==( COMPLETE_TYPE const& other ) const
	{
		return m_data == other.m_data;
	}

	bool operator==( T const& other ) const
	{
		return m_data == other;
	}

	bool operator!=( COMPLETE_TYPE const& other ) const
	{
		return m_data != other.m_data;
	}

	bool operator!=( T const& other ) const
	{
		return m_data != other;
	}

	const T get( void ) const
	{
		return m_data;
	}
};

template
<
	typename T,
	T INVALID_VAL,
	void(*DTOR)(T)
>
class unique_res : public weak_res<T,INVALID_VAL>
{
	typedef unique_res<T,INVALID_VAL,DTOR> UNIQUE_TYPE;
public:
	unique_res( UNIQUE_TYPE const& ) = delete;
	UNIQUE_TYPE& operator=( UNIQUE_TYPE const& other ) = delete;

	unique_res( void )
	{
		this->m_data = INVALID_VAL;
	}

	unique_res( UNIQUE_TYPE && other )
	{
		this->m_data = other.m_data;
		other.m_data = INVALID_VAL;
	}

	unique_res( T && src )
	{
		this->m_data = src;
		src = INVALID_VAL;
	}

	UNIQUE_TYPE& operator=( UNIQUE_TYPE && other )
	{
		clear();
		std::swap( this->m_data, other.m_data );
		return *this;
	}

	UNIQUE_TYPE& operator=( T && other )
	{
		clear();
		std::swap( this->m_data, other );
		return *this;
	}

	T release( void )
	{
		T ret = this->m_data;
		this->m_data = INVALID_VAL;
		return ret;
	}

	void clear( void )
	{
		if( INVALID_VAL != this->m_data )
		{
			DTOR( this->m_data );
			this->m_data = INVALID_VAL;
		}
	}

	~unique_res( void )
	{
		clear();
	}
};

#endif
