// Copyright (c) 2020 Morel Bérenger
// 
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.
// 
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
// 
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would be
//    appreciated but is not required.
// 2. Altered source versions must be plainly marked as such, and must not be
//    misrepresented as being the original software.
// 3. This notice may not be removed or altered from any source distribution.

// option parser: implementation

#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <errno.h>
#include <limits.h>

#include "optparser.hpp"
#include "utils.hpp"

// those strings must be defined in same order than the enum
// Trick: use vim's block insertion to check them
char const *parse_error_msgs[] =
{
	/*NONE       */"no error",
	/*IGNORED    */"not an option",
	/*MAX_COUNT  */"count overflow",
	/*SET_NO_VAL */"this option needs a value",
	/*SET_VAL_IGN*/"this option does not needs a value",
	/*SET_FAIL   */"failed to parse the value",
	/*BAD_ARGS   */"application bug: bad arguments to call",
	/*BAD_SETTER */"application bug: bad values in option",
};

parse_error_t parse_cmd_opt( char * arg, opt_desc_t* start, opt_desc_t const* end )
{
	BUG_CHECK( !( arg && start && end && end > start ), BAD_ARGS );

	if( arg[0] != '-' || arg[1] == 0 )
	{
		return IGNORED;
	}

	opt_desc_t* opt = start;
	char* argv = arg + 2;

	//search for option (stores in opt)
	if( arg[1] == '-' ) //long
	{
		char* name = argv;
		size_t name_sz;
		argv = strchr( name, '=' );
		if( argv )
		{
			assert( argv > name );
			name_sz = static_cast<size_t>( argv - name );
			++argv;
		}
		else
		{
			name_sz = strlen( name );
		}

		for( ; true
				&& opt != end
				&& 0 != strncmp( name, opt->option, name_sz )
				&& name_sz != strlen( opt->option ); ++opt )
		{
		}
	}
	else //short
	{
		uint32_t name = static_cast<uint8_t>( arg[1] );
		if( name & 0x80 )
		{
			assert( !( name & 0x80 ) && "TODO: utf-8 support" );
			abort();
		}
		for( ; opt != end && opt->short_option != name; ++opt )
		{
		}
	}

	//BUG_CHECK( ( opt->set && !opt->value ) || ( !opt->set && opt->value ), BAD_SETTER );

	if( opt == end )
	{
		return IGNORED;
	}

	if( opt->value && empty_array( argv ) )
	{
		return SET_NO_VAL;
	}

	if( !opt->value && !empty_array( argv ) )
	{
		return SET_VAL_IGN;
	}

	if( opt->set && opt->set( opt->value, argv ) )
	{
		return SET_FAIL;
	}

	if( opt->count < UINT32_MAX )
	{
		++opt->count;
	}
	return opt->count == UINT32_MAX ? MAX_COUNT : NONE;
}

void print_opts( FILE_PTR target, opt_desc_t const* start, opt_desc_t const* end )
{
	char short_name[] = ",-????";
	for( ; start != end; ++start )
	{
		memcpy( short_name + 2, &start->short_option, sizeof( start->short_option ) );
		fprintf( target, "\t* --%s%s: %s",
				start->option,
				start->short_option ? short_name : "",
				start->description
				);
		//fail to show current val
		if( start->show && start->show( start->value, target ) )
		{
			abort();
		}
		fputc( '\n', target );
	}
}
