// Copyright (c) 2020 Morel Bérenger
// 
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.
// 
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
// 
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would be
//    appreciated but is not required.
// 2. Altered source versions must be plainly marked as such, and must not be
//    misrepresented as being the original software.
// 3. This notice may not be removed or altered from any source distribution.

#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <assert.h>
#include <string.h>

#include <iterator>
#include <algorithm>

#include "utils.hpp"

#include <sys/ioctl.h>
#include <termios.h>

//would probably be better without that many ifs
int esc_fputs( const char *s, FILE *stream )
{
	assert( s && stream && "nullpointer provided to esc_fputs" );
	if( !( s && stream ) )
	{
		errno = EINVAL;
		return EOF;
	}

	const char *str_end = s + strlen( s );
	const char *chunk_start = s;
	const char *chunk_end;
	char bslash[2] = { '\\', '\\' };
	char tohex [4] = { '\\', 'x', 0, 0 };
	char ch;
	char const* buf;
	size_t sz;

	do
	{
		chunk_end = std::find_if( chunk_start, str_end, []( const char c ){ return c < 0x20 || c == '\\'; } );
		assert( chunk_end - chunk_start >= 0 );
		sz = static_cast<size_t>( chunk_end - chunk_start );
		if( sz != 0 && !fwrite( chunk_start, sz, 1, stream ) )
		{
			return INT_MAX;
		}
		if( chunk_end != str_end )
		{
			ch = *chunk_end;
			buf = bslash;
			sz = sizeof( bslash );
			if( ch != '\\' )
			{
				buf = tohex;
				sz = sizeof( tohex );
				//quick'n dirty ascii to hex convertion for values < 0x20 (space)
				tohex[2] = '0' + ( ch >> 4 );
				ch &= 0x0F;
				tohex[3] = ch + '0' + ( ch > 9 ? 'A' - ( '9' + 1 ) : 0 );
			}
			if( !fwrite( buf, sz, 1, stream ) )
			{
				return INT_MAX;
			}
			++chunk_end;
		}
		chunk_start = chunk_end;
	} while( chunk_start != str_end );
	return 0;
}

size_t ungets( char const* const str, size_t str_sz )
{
	// reverse push, since otherwise one would start with EOF
	for( size_t i = str_sz; i; --i )
	{
		if( EOF == ungetc( str[i-1], stdin ) )
		{
			return i;
		}
	}
	return 0;
}

bool term_ch_size( uint16_t *w, uint16_t *h, int fd )
{
	struct winsize wsz;
	if( -1 == ioctl( fd, TIOCGWINSZ, &wsz ) )
	{
		return true;
	}
	*w = wsz.ws_col;
	*h = wsz.ws_row;
	return false;
}

bool indent_txt( char const* txt, char const* end, uint8_t level, uint16_t max_width, uint8_t tab_width, FILE* target )
{
	uint16_t indent_size = tab_width * level;
	//bad args
	if( !txt || !target )
	{
		return true;
	}
	//overflow
	if( indent_size < tab_width || indent_size < level )
	{
		return true;
	}
	//indent wider than screen
	if( indent_size > max_width )
	{
		return true;
	}
	max_width -= indent_size;
	assert( max_width > 0 );

	uint16_t pos        = 0;
	uint16_t last_blank = 0;
	char const* ptr = txt;
	for( ; ptr != end; ++ptr, ++pos )
	{
		if( isblank( *ptr ) )
		{
			last_blank = pos;
		}

		if( *ptr == '\n' || pos == max_width )
		{
			if( *ptr != '\n' && last_blank != 0 )
			{
				assert( pos >= last_blank );
				ptr -= pos - last_blank;
				pos = last_blank;
			}
			for( uint8_t l = 0; l < level; ++l )
			{
				fputc( '\t', target );
			}
			fprintf( target, "%.*s", pos, txt );
			last_blank = 0;
			pos = 0;
			txt = ptr + 1;
			if( *ptr != '\n' )
			{
				fputc( '\n', target );
			}
		}
	}
	return false;
}

size_t file_size( char const* filename )
{
	FILE* f = fopen( filename );
	if( !f )
	{
		syserr( "file_size: failed to open" );
		return 0;
	}
	size_t ret = file_size( f );
	int err = errno;
	if( EOF == fclose( f ) )
	{
		syserr( "file_size: failed to close" );
		errno = err;
	}
	return ret;
}

size_t file_size( FILE* file )
{
	int err;
	assert( file != nullptr );
	long old = ftell( file );
	if( -1 == old )
	{
		err = errno;
		syserr( "1st ftell in file_size( FILE* ) failed" );
		errno = err;;
		return true;
	}

	if( -1 == fseek( file, 0, SEEK_END ) )
	{
		err = errno;
		syserr( "1st fseek in file_size( FILE* ) failed" );
		errno = err;;
		return true;
	}

	long size = ftell( file );
	if( -1 == old )
	{
		err = errno;
		syserr( "2nd ftell in file_size( FILE* ) failed" );
		errno = err;;
		return true;
	}
	if( -1 == fseek( file, old, SEEK_SET ) )
	{
		err = errno;
		syserr( "2nd fseek in file_size( FILE* ) failed" );
		errno = err;;
		return 0;
	}
	return size;
}
