// Copyright (c) 2020 Morel Bérenger
// 
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.
// 
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
// 
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would be
//    appreciated but is not required.
// 2. Altered source versions must be plainly marked as such, and must not be
//    misrepresented as being the original software.
// 3. This notice may not be removed or altered from any source distribution.

// a partial vector implementation.
// Why use it? Because:
// * it does *not* rely on exceptions;
// * it does *not* requires special tooling for debugging;
// * it allows lot smaller final binaries (than STL);
// You can just include it, and not use it by defining the
// WITH_STL flag: handy when you want STL for production,
// but this one for debugging.
//
// NOTE: the API is *not* 100% compatible, since it does not
// uses exceptions, but return values to signal errors, and
// is incomplete.
// Also, methods which are *not* in std::vector will be
// exposed in the end, to help handling memory.
//
// Bonus:
// * vector::remains() returns the number of unused slots
// * standalone "append" functions, which returning same
//   type as the underlying container's "push_back" method
// * standalone "real_size" function, returning amount of
//   memory used by a container. Does *not* count dynamic
//   memory allocated by contained objects.

#ifndef VECTOR_HPP
#define VECTOR_HPP

#ifndef WITH_STL

#ifndef NDEBUG
#include <stdio.h>
#endif

//this could *highly* probably be replaced with:
//void* new( void* p ) { return p; }
#include <new>

//methods return true on error
//public API should rely as much as possible on private API:
//no need to pile useless calls: this just makes debug harder
//code should never throw
//The class is marked final to avoid useless vtable.
template <typename T>
class vector final
{
	T* m_data = nullptr;
	size_t m_size = 0;
	size_t m_alloc = 0;

	bool allocate( size_t new_size );
	bool release( size_t new_size );
	bool construct( size_t new_size );
public:
	typedef T* iterator;
	typedef T const* const_iterator;
	typedef T value_type;

	bool empty( void ) const;
	size_t size( void ) const;
	size_t capacity( void ) const;
	T const& operator[]( size_t index ) const;
	T& operator[]( size_t index );
	bool shrink_to_fit( void );
	bool reserve( size_t size );
	bool resize( size_t size );
	bool assign( const_iterator start, const_iterator end );

	//default
	vector( void ) = default;
	//copy
	vector( vector<T> const& );
	vector<T>& operator=( vector<T> const& );
	//move
	vector( vector<T> && );
	vector<T>& operator=( vector<T> && );
	//destructor
	~vector( void );

	T const* begin( void ) const;
	T * begin( void );
	T const* end( void ) const;
	T * end( void );

	T const* rbegin( void ) const;
	T * rbegin( void );
	T const* rend( void ) const;
	T * rend( void );

	T const* data( void ) const;
	T * data( void );

	T const& front( void ) const;
	T & front( void );

	T const& back( void ) const;
	T & back( void );

	bool push_back( const T& value );
	bool push_back( T&& value );
	void pop_back( void );

	// non-standard methods

	// how much objects can still be stored before realloc
	size_t remains( void ) const;
};

template <typename T>
bool vector<T>::allocate( size_t new_size )
{
	T* tmp = static_cast<T*>( realloc( m_data, new_size * sizeof( T ) ) );
	if( tmp )
	{
		m_data  = tmp;
		m_alloc = new_size;
	}
	return !tmp;
}

template <typename T>
bool vector<T>::release( size_t new_size )
{
	if( new_size > m_size )
	{
		return true;
	}
	T* start = m_data + new_size;
	T* end   = m_data + m_size;
	for( ; start != end; ++start )
	{
		start->T::~T();
	}
	m_size = new_size;
	return false;
}

template <typename T>
bool vector<T>::construct( size_t new_size )
{
	if( new_size < m_size )
	{
		return true;
	}

	T* start = m_data + m_size;
	T* end   = m_data + new_size;
	for( ; start != end; ++start )
	{
		new ( start ) T();
	}
	m_size = new_size;
	return false;
}

template <typename T>
size_t vector<T>::capacity( void ) const
{
	return m_alloc;
}

template <typename T>
bool vector<T>::empty( void ) const
{
	return m_size == 0;
}

template <typename T>
size_t vector<T>::size( void ) const
{
	return m_size;
}

template <typename T>
T const& vector<T>::operator[]( size_t index ) const
{
	return m_data[index];
}

template <typename T>
T& vector<T>::operator[]( size_t index )
{
	return m_data[index];
}

template <typename T>
bool vector<T>::shrink_to_fit( void )
{
	return allocate( m_size );
}

template <typename T>
bool vector<T>::reserve( size_t size )
{
	if( m_alloc < size )
	{
		return allocate( size );
	}
	return false;
}

template <typename T>
bool vector<T>::resize( size_t size )
{
	if( size < m_size )
	{
		return release( size );
	}

	if( allocate( size ) )
	{
		return true;
	}
	return construct( size );
}

template <typename T>
bool vector<T>::assign( const_iterator start, const_iterator end )
{
	if( end - start < 0 )
	{
		return true;
	}
	size_t new_size = static_cast<size_t>( end - start );

	release( new_size );
	if( allocate( new_size ) )
	{
		return true;
	}
	m_size = new_size;

	T* dst = m_data;
	for( ; start != end; ++dst, ++start )
	{
		*dst = *start;
	}

	return false;
}

template <typename T>
vector<T>::vector( vector<T> const& other )
{
#ifndef NDEBUG
	fprintf( stderr, "vector copy ctor: from %p[%zu] to %p[%zu]\n",
	                 static_cast<void*>( other.m_data ), other.m_size,
	                 static_cast<void*>( m_data ), m_size
	);
#endif
	assign( other.begin(), other.end() );
}

template <typename T>
vector<T>& vector<T>::operator=( vector<T> const& other )
{
#ifndef NDEBUG
	fprintf( stderr, "vector copy assign: from %p[%zu] to %p[%zu]\n",
	                 static_cast<void*>( other.m_data ), other.m_size,
	                 static_cast<void*>( m_data ), m_size
	);
#endif
	assign( other.begin(), other.end() );
	return *this;
}

template <typename T>
vector<T>::vector( vector<T> && other )
{
	std::swap( other.m_data , m_data  );
	std::swap( other.m_size , m_size  );
	std::swap( other.m_alloc, m_alloc );
}

template <typename T>
vector<T>& vector<T>::operator=( vector<T> && other )
{
	std::swap( other.m_data , m_data  );
	std::swap( other.m_size , m_size  );
	std::swap( other.m_alloc, m_alloc );
	return *this;
}

template <typename T>
vector<T>::~vector( void )
{
	release( m_size );
	free( m_data );
}

template <typename T>
T const* vector<T>::data( void ) const
{
	return m_data;
}

template <typename T>
T * vector<T>::data( void )
{
	return m_data;
}

template <typename T>
T const* vector<T>::begin( void ) const
{
	return m_data;
}

template <typename T>
T * vector<T>::begin( void )
{
	return m_data;
}

template <typename T>
T const* vector<T>::rbegin( void ) const
{
	return end() - 1;
}

template <typename T>
T * vector<T>::rbegin( void )
{
	return end() - 1;
}

template <typename T>
T const* vector<T>::rend( void ) const
{
	return begin() - 1;
}

template <typename T>
T * vector<T>::rend( void )
{
	return begin() - 1;
}


template <typename T>
T const& vector<T>::front( void ) const
{
	return *m_data;
}

template <typename T>
T & vector<T>::front( void )
{
	return *m_data;
}


template <typename T>
T const* vector<T>::end( void ) const
{
	return m_data + m_size;
}

template <typename T>
T * vector<T>::end( void )
{
	return m_data + m_size;
}

template <typename T>
T const& vector<T>::back( void ) const
{
	return m_data[m_size - 1];
}

template <typename T>
T & vector<T>::back( void )
{
	return m_data[m_size - 1];
}

template <typename T>
bool vector<T>::push_back( const T& value )
{
	size_t new_size = m_size + 1;
	if( new_size >= capacity() && allocate( new_size ) )
	{
		return true;
	}
	m_data[m_size] = value;
	++m_size;
	return false;
}

template <typename T>
bool vector<T>::push_back( T&& value )
{
	size_t new_size = m_size + 1;
	if( new_size >= capacity() && allocate( new_size ) )
	{
		return true;
	}
	new (m_data + m_size ) T( std::move( value ) );
	//m_data[m_size] = std::move( value );
	++m_size;
	return false;
}

template <typename T>
void vector<T>::pop_back( void )
{
	if( m_size )
	{
		--m_size;
		m_data[m_size].~T();
	}
}

// non-standard methods

// how much objects can still be stored before realloc
template <typename T>
size_t vector<T>::remains( void ) const
{
	return capacity() - size();
}

#else
#include <vector>
using std::vector;
#endif

// append a single element to dst
template<typename C, typename T>
auto append( C& dst, T && src ) -> decltype( std::declval<C>().push_back( T() ) )
{
	return dst.push_back( src );
}

// append a single element to dst
template<typename C, typename T>
auto append( C& dst, T const& src ) -> decltype( std::declval<C>().push_back( T() ) )
{
	return dst.push_back( src );
}

// append a set to dst
template<typename C>
auto append( C& dst, typename C::const_iterator src_start, typename C::const_iterator src_end )
	-> decltype( std::declval<C>().push_back( typename C::value_type() ) )
{
	assert( src_end >= src_start );
	if( dst.reserve( dst.size() + static_cast<size_t>( src_end - src_start ) ) )
	{
		return true;
	}
	for( ; src_start != src_end; ++src_start )
	{
		if( dst.push_back( *src_start ) )
		{
			return true;
		}
	}
	return false;
}

// append a set to dst
template<typename C>
auto append( C& dst, C const& src )
	-> decltype( std::declval<C>().push_back( typename C::value_type() ) )
{
	if( dst.reserve( dst.size() + src.size() ) )
	{
		return true;
	}
	for( auto el : src )
	{
		dst.push_back( el );
	}
	return false;
}

// return the amount of memory used by current instance.
// Obviously, it does not takes into account memory
// allocated by contained objects.
template <typename T>
size_t real_size( vector<T> const& target )
{
	return sizeof( target ) + target.capacity() * sizeof( T );
}

#endif
